/*
 * Copyright (c) 2014 Atmark Techno, Inc.  All Rights Reserved.
 *  Tetsuhisa KOSEKI <koseki@atmark-techno.com>
 */

#include <autoconf.h>
#include <hermit.h>
#include <buffer.h>
#include <io.h>
#include <herrno.h>
#include <scan.h>
#include <flash.h>
#include <string.h>
#include <delay.h>
#include <spi.h>
#include <arch/system.h>
#include <arch/rcm2_qspi.h>

#define qspi_writeb(addr, val)	write8(addr, val)
#define qspi_writew(addr, val)	write16(addr, val)
#define qspi_writel(addr, val)	write32(addr, val)

#define qspi_readb(addr)	read8(addr)
#define qspi_readw(addr)	read16(addr)
#define qspi_readl(addr)	read32(addr)

static int qspi_send(u8 *tdata)
{
	while (!(qspi_readb(QSPI_SPSR) & QSPI_SPSR_SPTEF));

	qspi_writeb(QSPI_SPDR, *tdata);

	return 0;
}

static int qspi_recv(u8 *rdata)
{
	while (!(qspi_readb(QSPI_SPSR) & QSPI_SPSR_SPRFF));

	*rdata = qspi_readb(QSPI_SPDR);

	return 0;
}

static int qspi_xfer(u8 *tdata, u8 *rdata)
{
	int ret;

	ret = qspi_send(tdata);
	if (ret)
		return ret;

	return qspi_recv(rdata);
}

int spi_xfer_quad(unsigned int cmd_len, const void *cmd,
		  unsigned int data_len, const void *data_out,
		  void *data_in, u32 flags)
{
	u8 *tdata;
	u8 ddata = 0;
	int ret = 0;

	if ((data_out == NULL) && (data_in == NULL)) {
		if (flags & SPI_XFER_END)
			qspi_writeb(QSPI_SPCR, QSPI_SPCR_MSTR);
		return 0;
	}

	/* SPI Function Dsiable */
	qspi_writeb(QSPI_SPCR, QSPI_SPCR_MSTR);

	qspi_writew(QSPI_SPCMD0, QSPI_SPCMD_SCKDEN | QSPI_SPCMD_SLNDEN |
		    QSPI_SPCMD_SPNDEN | QSPI_SPCMD_SSLKP |
		    QSPI_SPCMD_BRDV_DIV2);
	qspi_writew(QSPI_SPCMD1, QSPI_SPCMD_SCKDEN | QSPI_SPCMD_SLNDEN |
		    QSPI_SPCMD_SPNDEN | QSPI_SPCMD_SSLKP |
		    QSPI_SPCMD_BRDV_DIV2);

	if (data_out != NULL)
		qspi_writew(QSPI_SPCMD2, QSPI_SPCMD_SCKDEN |
			    QSPI_SPCMD_SLNDEN | QSPI_SPCMD_SPNDEN |
			    QSPI_SPCMD_SPIMOD_QUAD | QSPI_SPCMD_BRDV_DIV2);
	else
		qspi_writew(QSPI_SPCMD2, QSPI_SPCMD_SCKDEN |
			    QSPI_SPCMD_SLNDEN | QSPI_SPCMD_SPNDEN |
			    QSPI_SPCMD_SPIMOD_QUAD | QSPI_SPCMD_SPRW |
			    QSPI_SPCMD_BRDV_DIV0 | QSPI_SPCMD_CPHA);

	qspi_writeb(QSPI_SPBFCR, QSPI_SPBFCR_TXRST | QSPI_SPBFCR_RXRST);
	qspi_writeb(QSPI_SPBFCR, 0x00);
	/* Sequence: 0->1->2->0 */
	qspi_writeb(QSPI_SPSCR, 0x02);

	qspi_writel(QSPI_SPBMUL0, 1);
	qspi_writel(QSPI_SPBMUL1, cmd_len - 1);
	qspi_writel(QSPI_SPBMUL2, data_len);

	/* SPI Function Enable */
	qspi_writeb(QSPI_SPCR, QSPI_SPCR_SPE | QSPI_SPCR_MSTR);

	/* command transfer */
	if (cmd != NULL)
		tdata = (u8 *)cmd;
	else
		return -H_EINVAL;

	while (cmd_len > 0) {
		ret = qspi_xfer(tdata, &ddata);
		if (ret)
			break;
		tdata++;
		cmd_len--;
	}

	/* data transfer */
	if ((data_out != NULL) && (data_in != NULL))
		hprintf("qspi_xfer_qread: full duplex is no supported\n");

	if (data_out != NULL) {
		tdata = (u8 *)data_out;

		while (data_len > 0) {
			ret = qspi_send(tdata);
			if (ret)
				break;
			tdata++;
			data_len--;
		}
		while (qspi_readw(QSPI_SPBDCR) & QSPI_SPBDCR_TXBC_MASK);
	} else if (data_in != NULL) {
		tdata = (u8 *)data_in;

		while (data_len > 0) {
			ret = qspi_recv(tdata);
			if (ret)
				break;
			tdata++;
			data_len--;
		}
	}

	return ret;
}


int spi_xfer_normal(unsigned int data_len, const void *data_out,
		    void *data_in, unsigned long flags)
{
	u8 *tdata, *rdata, dtdata = 0, drdata;
	int ret = 0;

	if ((data_out == NULL) && (data_in == NULL)) {
		if (flags & SPI_XFER_END)
			qspi_writeb(QSPI_SPCR, QSPI_SPCR_MSTR);
		return 0;
	}

	if (flags & SPI_XFER_BEGIN) {
		/* SPI Function Dsiable */
		qspi_writeb(QSPI_SPCR, QSPI_SPCR_MSTR);

		qspi_writew(QSPI_SPCMD0, QSPI_SPCMD_SCKDEN |
			    QSPI_SPCMD_SLNDEN | QSPI_SPCMD_SPNDEN |
			    QSPI_SPCMD_SSLKP | QSPI_SPCMD_BRDV_DIV2);

		qspi_writeb(QSPI_SPBFCR, QSPI_SPBFCR_TXRST |
			    QSPI_SPBFCR_RXRST);
		qspi_writeb(QSPI_SPBFCR, 0x00);

		/* Sequence: 0->1->0 */
		qspi_writeb(QSPI_SPSCR, 0x00);
	}

	qspi_writel(QSPI_SPBMUL0, data_len);

	/* SPI Function Enable */
	qspi_writeb(QSPI_SPCR, QSPI_SPCR_SPE | QSPI_SPCR_MSTR);

	if (data_out != NULL)
		tdata = (u8 *)data_out;
	else
		tdata = &dtdata;

	if (data_in != NULL)
		rdata = (u8 *)data_in;
	else
		rdata = &drdata;

	while (data_len > 0) {
		ret = qspi_xfer(tdata, rdata);
		if (ret)
			break;

		if (data_out != NULL)
			tdata++;

		if (data_in != NULL)
			rdata++;

		data_len--;
	}

	if (flags & SPI_XFER_END)
		qspi_writeb(QSPI_SPCR, QSPI_SPCR_MSTR);

	return ret;
}


int spi_initialize(void)
{
	/* SPI Function Dsiable */
	qspi_writeb(QSPI_SPCR, QSPI_SPCR_MSTR);

	qspi_writeb(QSPI_SSLP, QSPI_SSLP_LOW_ACTIVE);
	qspi_writeb(QSPI_SPPCR, QSPI_SPPCR_IO3FV | QSPI_SPPCR_IO2FV);

	/* Bit Rate: 24.38Mbps*/
	qspi_writeb(QSPI_SPBR, 0x02);

	qspi_writeb(QSPI_SPDCR, 0x00);

	/* Clock Delay: 1.5 SPCLK cycles(SPCMD.SCKDEN == 1) */
	qspi_writeb(QSPI_SPCKD, 0x00);
	/* SSL Negation Delay: 1 SPCLK cycles(SPCMD.SLNDEN == 1) */
	qspi_writeb(QSPI_SSLND, 0x00);
	/* Next-Access Delay: 1 SPCLK cycles(SPCMD.SPNDEN == 1) */
	qspi_writeb(QSPI_SPND, 0x00);

	/* Command setting */
	qspi_writew(QSPI_SPCMD0, QSPI_SPCMD_SCKDEN | QSPI_SPCMD_SLNDEN |
		    QSPI_SPCMD_SPNDEN | QSPI_SPCMD_SSLKP |
		    QSPI_SPCMD_BRDV_DIV2);
	qspi_writew(QSPI_SPCMD0, QSPI_SPCMD_SCKDEN | QSPI_SPCMD_SSLKP |
		    QSPI_SPCMD_BRDV_DIV2);

	/* Transmit/Receive Data Reset */
	qspi_writeb(QSPI_SPBFCR, QSPI_SPBFCR_TXRST | QSPI_SPBFCR_RXRST);
	qspi_writeb(QSPI_SPBFCR, QSPI_SPBFCR_TXTRG_0B);

	/* Sequence: 0->0->... */
	qspi_writeb(QSPI_SPSCR, 0x00);

	/* SPI Function Enable */
	qspi_writeb(QSPI_SPCR, QSPI_SPCR_SPE | QSPI_SPCR_MSTR);

	return 0;
}

void spi_start_timer(u32 usecs)
{
	u32 count;
	u8 tstr;

	/* SUB = 10MHz, TCNT_CLK = 2.5MHz */
	count = (2 * usecs) + ((usecs + 1) / 2);

	write16(TMU(0, TCR), 0);
	write32(TMU(0, TCOR), count);
	write32(TMU(0, TCNT), count);

	tstr = read8(TMU_TSTR0);
	write8(TMU_TSTR0, tstr | 0x1);
}

void spi_stop_timer(void)
{
	u8 tstr;

	write16(TMU(0, TCR), 0x0000);
	tstr = read8(TMU_TSTR0);
	write8(TMU_TSTR0, tstr & ~0x1);
}

int spi_check_timer(void)
{
	if (!(read16(TMU(0, TCR)) & 0x100))
		return -1;

	return 0;
}
